@@ -236,6 +236,7 @@ pub enum SyntaxError {
     TS1245,
     TS1267,
     TS1273(JsWord),
+    TS1274(JsWord),
     TS1383,
     TS2206,
     TS2207,
@@ -584,6 +585,12 @@ impl SyntaxError {
             SyntaxError::TS1273(word) => {
                 format!("'{}' modifier cannot appear on a type parameter", word).into()
             }
+            SyntaxError::TS1274(word) => format!(
+                "'{}' modifier can only appear on a type parameter of a class, interface or type \
+                 alias",
+                word
+            )
+            .into(),
             SyntaxError::TS1383 => "Only named exports may use 'export type'.".into(),
             SyntaxError::TS2206 => "The 'type' modifier cannot be used on a named import when \
                                     'import type' is used on its import statement."
